module Cross_Deadlock_VC where

import Datastructures
import Data.Sequence (fromList)
import Control.Exception (assert)
import qualified Data.Sequence as Seq


import Debug.Trace

-- build a merge with $n$ inputs from normal merges
n_merge n idx trgt = assert (n >= 2)
			map (\i -> (Merge, idx ++ show i, [Nothing,Nothing], (if i == n - 2 then [trgt] else [Just(idx ++ (show $ i + 1),0)])))
				[0..n-2]

data Color = GetX | FwdGetX | Data | PutX | WriteBackNack | WriteBackAck | Invalidate | ReplEvent | Load | InvEvent | ReplAck
  deriving (Show, Eq, Ord, Enum)
data Packet = Token | Packet {
	x :: Int,
	y :: Int,
	x_src :: Int,
	y_src :: Int,
	color :: Color
} deriving (Show, Eq, Ord)

allPackets = [Token] ++
			 map (\(x,y,x',y',c) -> Packet x y x' y' c) [(x,y,x',y',c) | x <- [0,1], y <- [0,1], x' <- [0,1], y' <- [0,1], c <- [GetX ..]]


show_p Token = "token"
show_p (Packet x y x' y' c) = "x" ++ show x ++ "_y" ++ show y ++ "xx" ++ show x' ++ "_yy" ++ show y' ++ "_c" ++ show c

isToken Token = True
isToken _ = False
isLoad Load = True
isLoad _ = False
isWBN (Packet x y x' y' WriteBackNack) = True
isWBN _ = False
isWBA (Packet x y x' y' WriteBackAck) = True
isWBA _ = False
isPutX (Packet x y x' y' PutX) = True
isPutX _ = False
isFwdGetX (Packet x y _ _ FwdGetX) x' y' = (x,y) == (x',y')
isFwdGetX _ _ _ = False
isInv (Packet x y x' y' Invalidate) = True
isInv _ = False
isData (Packet x y _ _ Data) x' y' = (x,y) == (x',y')
isData _ _ _ = False
isGetX (Packet x y x' y' GetX) = True
isGetX _ = False

intoToken d = Token



to_local x' y' = Switch (\d -> x d == x' && y d == y') ""
to_x     x' y' = Switch (\d -> x d /= x') ""
to_N     x' y' = Switch (\d -> y d < y') ""
to_W     x' y' = Switch (\d -> x d < x') ""


qsize = 3

add_vc n (c,id,ins,outs) =
	(c,id++show n,map add_vc_to_channel ins, map add_vc_to_channel outs)
	where
		add_vc_to_channel Nothing = Nothing
		add_vc_to_channel (Just (idx,p)) = Just (idx ++ show n, p) 

xy_node_p x y n =
	map (add_vc n)
		[
			(Queue qsize,			"N",				[Nothing],			[Just("switchLN",0)]),
			(Queue qsize,			"S",				[Nothing],			[Just("switchLS",0)]),
			(Queue qsize,			"E",				[Nothing],			[Just("switchLEV",0)]),
			(Queue qsize,			"W",				[Nothing],			[Just("switchLWV",0)]),
			(to_local x y,			"switchLN",			[Nothing],			[Just("mergeO",0), Just("mergeN",0)]),
			(to_local x y,			"switchLS",			[Nothing],			[Just("mergeO",1), Just("mergeS",0)]),
			(to_local x y,			"switchLEV",		[Nothing],			[Just("mergeO",2), Just("switchEV",0)]),
			(to_local x y,			"switchLWV",		[Nothing],			[Just("mergeO",3), Just("switchWV",0)]),
			(to_x x y,				"switchEV",			[Nothing],			[Just("mergeE",1), Just("switchENS",0)]),
			(to_x x y,				"switchWV",			[Nothing],			[Just("mergeW",1), Just("switchWNS",0)]),
			(to_N x y,				"switchENS",		[Nothing],			[Just("mergeN",1), Just("mergeS",1)]),
			(to_N x y,				"switchWNS",		[Nothing],			[Just("mergeN",2), Just("mergeS",2)]),
			(to_local x y,			"switchToFabric0",	[Nothing],			[Just("mergeO",4), Just("switchToFabric1",0)]),
			(to_x x y,				"switchToFabric1",	[Nothing],			[Just("switchWE",0), Just("switchNS",0)]),
			(to_W x y,				"switchWE",			[Nothing],			[Just("mergeW",0), Just("mergeE",0)]),
			(to_N x y,				"switchNS",			[Nothing],			[Just("mergeN",3), Just("mergeS",3)]),
			(Merge,					"mergeE",			[Nothing,Nothing],	[Just("outE", 0)]),
			(Merge,					"mergeW",			[Nothing,Nothing],	[Just("outW", 0)])
		]



natDecr :: Int -> Int
natDecr n = if n == 0 then 0 else n - 1


from_local_to_vc_getX x' y' = Switch isGetX ""
from_local_to_vc_putX x' y' = Switch isPutX ""

xy_node x y = Hierarchy (fromList $
							(concatMap (\n -> [Flat $ build $ xy_node_p x y n]
												++
											    [Flat $ build $ n_merge 5 ("mergeO"++show n) (Just("arrive"++show n,0))]
												++
											    [Flat $ build $ n_merge 4 ("mergeN"++show n) (Just("outN"++show n,0))]
												++
											    [Flat $ build $ n_merge 4 ("mergeS"++show n) (Just("outS"++show n,0))])
										[0..2])
							++
							[Flat $ build $ 
								[(from_local_to_vc_getX x y, "switchIn", [Nothing], [Just ("to_switchIn0", 0), Just ("to_vc_switch_putX", 0)]), 
								 (from_local_to_vc_putX x y, "to_vc_switch_putX", [Nothing], [Just ("to_switchIn1", 0), Just ("to_switchIn2", 0)]),
								 (Merge, 					 "mrg_arrive1", 	  [Nothing, Nothing], [Just ("mrg_arrive2", 0)]),
								 (Merge, 					 "mrg_arrive2", 	  [Nothing, Nothing], [Just ("arrive", 0)])]
							]
						)
						(\n s p -> case (n,s) of
									(0,"mergeO0") -> (1,  "mergeO0" ++ show (natDecr p), if p == 0 then 0 else 1)
									(0,"mergeN0") -> (2,  "mergeN0" ++ show (natDecr p), if p == 0 then 0 else 1)
									(0,"mergeS0") -> (3,  "mergeS0" ++ show (natDecr p), if p == 0 then 0 else 1)
									(4,"mergeO1") -> (5,  "mergeO1" ++ show (natDecr p), if p == 0 then 0 else 1)
									(4,"mergeN1") -> (6,  "mergeN1" ++ show (natDecr p), if p == 0 then 0 else 1)
									(4,"mergeS1") -> (7,  "mergeS1" ++ show (natDecr p), if p == 0 then 0 else 1)
									(8,"mergeO2") -> (9,  "mergeO2" ++ show (natDecr p), if p == 0 then 0 else 1)
									(8,"mergeN2") -> (10, "mergeN2" ++ show (natDecr p), if p == 0 then 0 else 1)
									(8,"mergeS2") -> (11, "mergeS2" ++ show (natDecr p), if p == 0 then 0 else 1)
									(1,"arrive0") -> (12, "mrg_arrive1", 0)
									(5,"arrive1") -> (12, "mrg_arrive1", 1)
									(9,"arrive2") -> (12, "mrg_arrive2", 1)
									(12,"to_switchIn0") -> (0, "switchToFabric00", 0)
									(12,"to_switchIn1") -> (4, "switchToFabric01", 0)
									(12,"to_switchIn2") -> (8, "switchToFabric02", 0)
									_ -> (n,s,p))

dimx = 4
dimy = 4

dirx = 3
diry = 3


num_of_caches = dimx*dimy-1
dir_number 	  = dirx + dimx*diry
cache_numbers = filter (/= dir_number) [0..dimx*dimy-1]

insert_token = Source [Token] ""

cache x y = [
			(Automaton ["I","M", "MI"] Token [(0,1,\p d-> p==0 && isToken d,	\p d -> Just (0, Packet dirx diry x y GetX)),
											  (1,2,\p d-> p==0 && isToken d,	\p d -> Just (0, Packet dirx diry x y PutX)),
											  (1,2,\p d-> p==1 && isInv d,		\p d -> Just (0, Packet dirx diry x y PutX)),
											  (2,2,\p d-> p==1 && isInv d,		\p d -> Nothing),
											  (2,0,\p d-> p==1 && isWBA d,		\p d -> Nothing)
										],
						"cache", [Nothing,Nothing], [Just ("injectQ", 0)]),
			(insert_token, "src", [], [Just("cache",0)]),
			(Function id "", "injectQ", [Nothing], [Just("inject", 0)]),
			(Function id "", "arriveQ", [Nothing], [Just("cache",1)])
	]


directory = [
			(Automaton
				(["I"] ++ (map (\cn -> "M" ++ show cn) cache_numbers) ++ (map (\cn -> "MI" ++ show cn) cache_numbers))
				Token
				(concatMap
					(\(n,cn) -> let (x,y) = (cn `mod` dimx, cn `div` dimx) in
						[	(0,n+1,					\p d -> p==1 && d == Packet dirx diry x y GetX,	\p d -> Nothing),
							(n+1,num_of_caches+n+1,	\p d -> p==0,									\p d -> Just (0, Packet x y dirx diry Invalidate)),
							(num_of_caches+n+1,0,	\p d -> p==1 && d == Packet dirx diry x y PutX, \p d -> Just (0, Packet x y dirx diry WriteBackAck)),
							(n+1,0,					\p d -> p==1 && d == Packet dirx diry x y PutX,	\p (Packet x y x_src y_src PutX) -> Just (0, Packet x_src y_src x y WriteBackAck))
						])
					$ zip [0..] cache_numbers),
				"dir", [Nothing,Nothing], [Just ("injectQ", 0)]),
			(insert_token, "src", [], [Just("dir",0)]),
			(Function id "", "injectQ", [Nothing], [Just("inject", 0)]),
			(Function id "", "arriveQ", [Nothing], [Just("dir",1)])
	]



{-
	Each router has three virtual channels: getX, putX, and the rest.
-}
hnetwork :: HNetwork Packet
hnetwork = Hierarchy (Seq.fromList $
					  (map (\n -> xy_node (n `mod` dimx) (n `div` dimx)) [0..dimx*dimy-1])
						++
					  (map (\n -> if n == dir_number then
									Flat $ build $ directory
								  else
									Flat $ build $ cache (n `mod` dimx) (n `div` dimx)) [0..dimx*dimy-1])
						)
					  (\n s p -> case s of
							   "outE0" -> if n `mod` dimx == dimx-1 then (n,"snk",0) else (n+1, "E0",0)
							   "outW0" -> if n `mod` dimx == 0 then (n,"snk",0) else (n-1, "W0",0)
							   "outN0" -> if n `div` dimx == 0 then (n,"snk",0) else (n-dimx, "N0",0)
							   "outS0" -> if n `div` dimx == dimy-1 then (n,"snk",0) else (n+dimx, "S0",0)
							   "outE1" -> if n `mod` dimx == dimx-1 then (n,"snk",0) else (n+1, "E1",0)
							   "outW1" -> if n `mod` dimx == 0 then (n,"snk",0) else (n-1, "W1",0)
							   "outN1" -> if n `div` dimx == 0 then (n,"snk",0) else (n-dimx, "N1",0)
							   "outS1" -> if n `div` dimx == dimy-1 then (n,"snk",0) else (n+dimx, "S1",0)
							   "outE2" -> if n `mod` dimx == dimx-1 then (n,"snk",0) else (n+1, "E2",0)
							   "outW2" -> if n `mod` dimx == 0 then (n,"snk",0) else (n-1, "W2",0)
							   "outN2" -> if n `div` dimx == 0 then (n,"snk",0) else (n-dimx, "N2",0)
							   "outS2" -> if n `div` dimx == dimy-1 then (n,"snk",0) else (n+dimx, "S2",0)
							   "arrive" -> (n+dimx*dimy, "arriveQ",0)
							   "inject" -> (n-dimx*dimy, "switchIn",0)
							   _      -> (n,s,p))




