module DeadlockDetection where
import Datastructures
import Control.Exception (assert)
import qualified Data.Sequence as Seq
import Debug.Trace
import qualified Data.Map as Map
import qualified Data.Set as Set
import Data.Data
import Data.Maybe
import Data.List
import Data.Foldable (toList)


import System.IO
import System.Process
import System.Environment
import System.Exit
import SMT
import System.IO.Unsafe (unsafePerformIO)




other 0 = 1
other 1 = 0



allTheSame :: (Eq a) => Seq.Seq a -> Bool
allTheSame xs = Seq.foldrWithIndex (\n a b -> b && a == Seq.index xs 0) (Seq.length xs /= 0) xs


-- TODO: use of equivalance sets

subsumes :: (Eq p_t, Ord p_t) => Formula p_t -> Formula p_t -> Bool
subsumes (Lit l) (OR fs)  = Set.member (Lit l) fs
subsumes (OR fs) (Lit l)  = fs == Set.singleton (Lit l)
subsumes (Lit l) (AND fs) = fs == Set.singleton (Lit l)
subsumes (AND fs) (Lit l) = Set.member (Lit l) fs
subsumes (AND fs1) (AND fs2) = fs2 `Set.isSubsetOf` fs1
subsumes (OR fs1) (OR fs2) = fs1 `Set.isSubsetOf` fs2
subsumes F _ = True
subsumes _ _ = False

remove_subsumptions_in_OR :: Show p_t => Ord p_t => [Formula p_t] -> [Formula p_t]
remove_subsumptions_in_OR [] = []
remove_subsumptions_in_OR (f:fs) = case find (\f' -> subsumes f f') fs of
									Just f' -> remove_subsumptions_in_OR fs
									Nothing -> f:(remove_subsumptions_in_OR fs)

remove_subsumptions_in_AND :: Show p_t => Ord p_t => [Formula p_t] -> [Formula p_t]
remove_subsumptions_in_AND [] = []
remove_subsumptions_in_AND (f:fs) = case find (\f' -> subsumes f' f) fs of
									Just f' -> remove_subsumptions_in_AND fs
									Nothing -> f:(remove_subsumptions_in_AND fs)

simplify_subsumptions :: Show p_t => Ord p_t => Formula p_t -> Formula p_t
simplify_subsumptions (OR fs) = let fs'  = Set.map simplify_subsumptions fs;
									fs'' = remove_subsumptions_in_OR (Set.toList fs') in
									if Data.List.length fs'' == 1 then
										fs''!!0
									else
										OR (Set.fromList fs'')
simplify_subsumptions (AND fs) = let fs'  = Set.map simplify_subsumptions fs;
									 fs'' = remove_subsumptions_in_AND (Set.toList fs') in
									if Data.List.length fs'' == 1 then
										fs''!!0
									else
										AND (Set.fromList fs'')
simplify_subsumptions f = f
								  


simplify_unrealism :: Ord p_t => Set.Set (Literal p_t)  -> Formula p_t -> Formula p_t
simplify_unrealism vis (OR fs) = let fs' = Set.filter (/= F) $ Set.map (simplify_unrealism vis) fs in
									if Set.size fs' == 1 then
										Set.findMin fs'
									else
										OR fs'
simplify_unrealism vis (AND fs) = let vis' = Set.union (Set.map takeLiteral $ Set.filter isLiteral fs) vis;
									  fs' = Set.map (simplify_unrealism vis') fs in
									if Set.size fs' == 1 then
										Set.findMin fs'
									else if F `Set.member` fs' then
										F
									else
										AND fs'
	where
		isLiteral (Lit l) = True
		isLiteral _ = False
		takeLiteral (Lit l) = l
simplify_unrealism vis (Lit (Is_Full q)) = if Is_Zero q Nothing `Set.member` vis then F else (Lit $ Is_Full q)
simplify_unrealism vis (Lit (Is_Zero q d)) = if (Is_Full q `Set.member` vis && d == Nothing) || Is_Gt_One q d `Set.member` vis then F else (Lit (Is_Zero q d))
simplify_unrealism vis (Lit (Is_Gt_One q d)) = if Is_Zero q d `Set.member` vis then F else (Lit (Is_Gt_One q d))
simplify_unrealism vis (Lit (Select m b)) = if (Select m $ not b) `Set.member` vis then F else (Lit (Select m b))
simplify_unrealism vis f = f


simplify_remove_selects (OR fs)  = let fs' = Set.filter (/= F) $ Set.map simplify_remove_selects fs in
									if Set.size fs' == 1 then Set.findMin fs' else OR fs'
simplify_remove_selects (AND fs) = let fs' = Set.filter (/= T) $ Set.map simplify_remove_selects fs in
									if Set.size fs' == 1 then Set.findMin fs' else AND fs'
simplify_remove_selects (Lit (Select m b)) = T
simplify_remove_selects f = f
							


unfold_formula x qs live vars nfqs show_p_t invs (OR fs) unfolding =
	Set.fold (unfold_formula x qs live vars nfqs show_p_t invs) unfolding fs 
unfold_formula x qs live vars nfqs show_p_t invs (AND fs) unfolding =
	Set.fold (unfold_formula x qs live vars nfqs show_p_t invs) unfolding fs 
unfold_formula x qs live vars nfqs show_p_t invs (Lit (Block i p d)) unfolding =
	case Map.lookup (Block i p d) unfolding of
		Just f -> unfolding
		Nothing -> let f = block x qs (i,p) d True live vars nfqs show_p_t invs;
					   unfolding' = Map.insert (Block i p d) f unfolding in
						unfold_formula x qs live vars nfqs show_p_t invs f unfolding'
unfold_formula x qs live vars nfqs show_p_t invs (Lit (Idle i p d)) unfolding =
	case Map.lookup (Idle i p d) unfolding of
		Just f -> unfolding
		Nothing -> let f = idle x qs (i,p) d True live vars nfqs show_p_t invs;
					   unfolding' = Map.insert (Idle i p d) f unfolding in
						unfold_formula x qs live vars nfqs show_p_t invs f unfolding'
unfold_formula x qs live vars nfqs show_p_t invs _ unfolding =
	unfolding




-- block :: (Show p_t, Ord p_t, Eq p_t) => Network p_t -> Set.Set Int -> (Int, Int) -> Maybe p_t -> Set.Set (Literal p_t) -> Map.Map Int Bool -> Seq.Seq Int -> Formula p_t
block x qs (i,p) d firstcall live vars nfqs show_p_t invs =
	if Seq.elemIndexL i live /= Nothing then F else
	let c = Seq.index x i in
	--trace ("DL: " ++ ide c ++ "(" ++ show p ++ ")." ++ show d ) $
	case prim c of 
		Queue s 	->	if not firstcall then
							if i `elem` nfqs then
								F
							else
								conjunct (Lit $ Is_Full i) (Lit $ Block i 0 Nothing)
						else 
							if Set.notMember i qs then
								block x qs (trgt c 0) Nothing False live vars nfqs show_p_t invs
							else
								let fs = Seq.mapWithIndex (\n color -> block x qs (trgt c 0) (Just color) False live vars nfqs show_p_t invs)
										  			      (Seq.index (colors c) 0) in
									if allTheSame fs then
										Seq.index fs 0
									else
										Seq.foldrWithIndex (\n color -> disjunct (conjunct (Lit $ Is_Gt_One i (Just color)) (Seq.index fs n))) F (Seq.index (colors c) 0)
		Fork		-> disjunct (block x qs (trgt c 0) d firstcall live vars nfqs show_p_t invs)
								(block x qs (trgt c 1) d firstcall live vars nfqs show_p_t invs)
		Join di		-> let d' = if di == p then d else Nothing in
						disjunct (block x qs (trgt c 0) d' firstcall live vars nfqs show_p_t invs)
								 (idle  x qs (init c $ other p) Nothing firstcall live vars nfqs show_p_t invs)
		Switch f s	-> case d of
						Just d' -> block x qs (trgt c (if f d' then 0 else 1)) d firstcall live vars nfqs show_p_t invs
						Nothing -> disjunct (if Seq.length (trgt_colors c 0) == 0 then F else block x qs (trgt c 0) d firstcall live vars nfqs show_p_t invs)
											(if Seq.length (trgt_colors c 1) == 0 then F else block x qs (trgt c 1) d firstcall live vars nfqs show_p_t invs)
		Merge		-> disjunct (
							 --trace (ide c ++ " select(1) " ++ if p==0 then "up" else "down" ++ " in") $
								conjunct (Lit $ Select i (p==0)) $
									case d of
										Nothing -> Seq.foldrWithIndex (\n color -> disjunct (block x qs (trgt c 0) (Just color) firstcall live vars nfqs show_p_t invs))
														 				F (Seq.index (colors c) p)
										Just d' -> block x qs (trgt c 0) d firstcall live vars nfqs show_p_t invs
						) $
							--trace (ide c ++ " select(2)  " ++ if p==0 then "down" else "up" ++ " in") $
								conjunct (Lit $ Select i (p==1)) $
								Seq.foldrWithIndex (\n color -> disjunct (conjunct	(block x qs (trgt c 0) (Just color) firstcall live vars nfqs show_p_t invs)
																					(block_bw x qs (init c $ other p) (Just color) live vars nfqs show_p_t invs)))
																F (Seq.index (colors c) $ other p)
		Sink dead	  -> fromBool dead
		Function f s  -> case d of
							Just d' -> block x qs (trgt c 0) (Just $ f d') firstcall live vars nfqs show_p_t invs
							Nothing -> block x qs (trgt c 0) d firstcall live vars nfqs show_p_t invs
		Source injs s -> if injs /= [] then block x qs (trgt c 0) Nothing False live vars nfqs show_p_t invs else T
		Automaton ss token ts -> if all (eventIsFalse c p d) ts then
										T
								 else
									if firstcall then
									-- The automaton is in a deadlock OR no transition fires with this packet at this channel 
										foldr (\s -> disjunct $ deadlockState x qs live vars nfqs show_p_t invs c i s ts) F [0..(length $ getStates ts)-1]
								 	else
										Lit $ Idle i 0 Nothing
	where trgt c p = case Seq.index (outs c) p of
						Just (Left i,p') -> (i,p')
	      init c p = case Seq.index (ins c) p of
						Just (Left i,p') -> (i,p')
	      trgt_colors c p = let (i',p') = trgt c p in
								Seq.index (colors (Seq.index x i')) p'
	      eventIsFalse c p d (s,s',e,f) = case d of { Just d -> not $ e p d; Nothing -> not $ eventCanBeTrue c e p}
	      eventCanBeTrue c e p = or (map (\d -> e p d) (toList $ Seq.index (colors c) p))








check_feasibility_unsafe x qs vars show_p_t invs f = 
	case f of
		T -> True
		F -> False
		_ -> unsafePerformIO $ do 
				let ret  = simplify_subsumptions $ simplify_remove_selects $ f
				let file = "temp.smt2"
				h <- openFile file WriteMode
				hPutStrLn h $ "(set-logic QF_LIA)\n" ++ invs
				let (smt',qs',vars') = export_formula_to_SMT x qs vars show_p_t Nothing f
				hPutStrLn h $ smt'
				hPutStrLn h $ "(check-sat)\n(get-model)"
				hClose h
				(exit,solver_output,err) <- (uncurry readProcessWithExitCode) ("cvc4", ["-m",file]) [] 
				case parseSMTOutput solver_output of
					Left err -> do
									print err
									exitFailure
					Right Nothing -> return False
					Right (Just model) -> return True


traceThis x = traceShow x x


deadlockState x qs live vars nfqs show_p_t invs c i s ts =
	--trace ("DS: " ++ ide c ++ "(" ++ show s ++ ").") $
 	conjunct (Lit $ InState i s) $ allTransitionsDead c $ filter (\(t,t',e,f) -> s == t) ts 
	where
	      allTransitionsDead c ts = foldr (conjunct . deadTransition c) T ts
		  -- Make a conjunction with for all in-channels and all colors of that channel the formula representing a dead transition
	      deadTransition c t = foldr (\p' f -> Seq.foldrWithIndex (\n color' -> conjunct (deadTransitionF c t p' color')) f (Seq.index (colors c) p')) 
									 T
									 [0..Seq.length (ins c)-1]
		  -- A transition is dead for in-channel p' with packet color', iff:
		  -- color' does not trigger the transition, or
		  -- the channel is idle for color', or
		  -- the produced packet is permanently blocked
	      deadTransitionF c (s,s',e,f) p' color' = if e p' color' then
														disjunct (case f p' color' of
																	Nothing -> F
																	Just (p'',color'') -> conjunct
																							(block_bw x qs (init c p') (Just color') live vars nfqs show_p_t invs)
																							(block x qs (trgt c p'') (Just color'') False live vars nfqs show_p_t invs))
																 (idle x qs (init c p') (Just color') False live vars nfqs show_p_t invs)
												   else T
	      trgt c p = case Seq.index (outs c) p of
						Just (Left i,p') -> (i,p')
	      init c p = case Seq.index (ins c) p of
						Just (Left i,p') -> (i,p')

									


--idle :: (Show p_t, Ord p_t, Eq p_t) => Network p_t -> Set.Set Int -> (Int, Int) -> Maybe p_t -> Set.Set  (Literal p_t) -> Map.Map Int Bool -> Seq.Seq Int -> Formula p_t
idle x qs (i,p) d firstcall live vars nfqs show_p_t invs =
	if Seq.elemIndexL i live /= Nothing then F else
	let c = Seq.index x i in
	--trace ("ID: " ++ ide c ++ "(" ++ show p ++ ")." ++ show d ) $
	case prim c of
		Queue s 	->	if (case d of
								Nothing -> Seq.length (Seq.index (colors c) 0) /= 0 
								Just d  -> Seq.elemIndexL d (Seq.index (colors c) 0) /= Nothing) then
							if not firstcall then
								Lit $ Idle i 0 d
							else if Seq.length (Seq.index (colors c) 0) == 1 then
									conjunct (Lit $ Is_Zero i Nothing)
											 (idle x qs (init c 0) Nothing False live vars nfqs show_p_t invs)
							else
								case d of
									Nothing -> conjunct (Lit $ Is_Zero i Nothing)
											 			(idle x qs (init c 0) Nothing False live vars nfqs show_p_t invs)
									Just d' -> -- TODO: assuming buffer semantics here
										let other_colors = Seq.filter (/= d') (Seq.index (colors c) 0);
											fs = Seq.mapWithIndex (\n color -> block x qs (trgt c 0) (Just color) False live vars nfqs show_p_t invs) other_colors;
											some_other_packet_blocked_at_head = if allTheSame fs then
																					Seq.index fs 0
																				else
																					Seq.foldrWithIndex (\n color -> disjunct (conjunct (Lit $ Is_Gt_One i (Just color)) (Seq.index fs n))) F other_colors in
											conjunct (Lit $ Is_Zero i d)
													 (disjunct (idle x qs (init c 0) d False live vars nfqs show_p_t invs)
													 		   (conjunct (Lit $ Is_Full i) some_other_packet_blocked_at_head))
	
						else
							T
		Fork		-> disjunct (idle x qs (init c 0) d firstcall live vars nfqs show_p_t invs)
								(block x qs (trgt c $ other p) d firstcall live vars nfqs show_p_t invs)
		Join di		-> disjunct (idle x qs (init c $ other di) Nothing firstcall live vars nfqs show_p_t invs)
								(idle x qs (init c $ di) d firstcall live vars nfqs show_p_t invs)
		Switch f s	-> case d of
						Nothing -> Seq.foldrWithIndex (\n color -> conjunct $ if f color == (p == 0) then idle x qs (init c 0) (Just color) firstcall live vars nfqs show_p_t invs else T) T (Seq.index (colors c) 0)
						Just d' -> if f d' /= (p == 0) then T else
									idle x qs (init c 0) d firstcall live vars nfqs show_p_t invs
		Merge		-> conjunct (if d == Nothing || Seq.elemIndexL (fromJust $ d) (Seq.index (colors c) 0) /= Nothing then idle x qs (init c 0) d firstcall live vars nfqs show_p_t invs else T)
							    (if d == Nothing || Seq.elemIndexL (fromJust $ d) (Seq.index (colors c) 1) /= Nothing then idle x qs (init c 1) d firstcall live vars nfqs show_p_t invs else T)
		Function f s  -> case d of
							Just d' -> if Seq.length (curr_colors c p) == 0 then T else
										Seq.foldrWithIndex (\n color -> disjunct $ if f color == d' then idle x qs (init c 0) (Just color) firstcall live vars nfqs show_p_t invs else F) F (Seq.index (colors c) 0)
							Nothing -> idle x qs (init c 0) d firstcall live vars nfqs show_p_t invs
		Source injs s -> case d of
							Nothing -> fromBool $ length injs == 0
							Just d  -> fromBool $ d `notElem` injs
		Automaton ss token ts -> if firstcall then
									-- The automaton is in a deadlock
									foldr (\s -> disjunct $ deadlockState x qs live vars nfqs show_p_t invs c i s ts) F [0..(length $ getStates ts)-1]
								 else
									Lit $ Idle i 0 Nothing
	where trgt c p = case Seq.index (outs c) p of
						Just (Left i,p') -> (i,p')
	      init c p = case Seq.index (ins c) p of
						Just (Left i,p') -> (i,p')
	      curr_colors c p = let (i',p') = trgt c p in
								Seq.index (colors (Seq.index x i')) p'




--block_bw :: (Show p_t, Ord p_t, Eq p_t) => Network p_t -> Set.Set Int -> (Int, Int) -> Maybe p_t -> Set.Set (Literal p_t) -> Map.Map Int Bool -> Seq.Seq Int -> Formula p_t
block_bw x qs (i,p) d live vars nfqs show_p_t invs =
	if Seq.elemIndexL i live /= Nothing then F else
	let c = Seq.index x i in
	--trace ("BL_BW: " ++ ide c ++ "(" ++ show p ++ ")." ++ show d ) $
	case prim c of 
		Queue s 	-> case d of 
							Just d' -> if Seq.elemIndexL d' (Seq.index (colors c) 0) /= Nothing then
										if Set.member i qs then Lit $ Is_Gt_One i d else Lit $ Is_Gt_One i Nothing
									   else F
							Nothing -> if Seq.length (Seq.index (colors c) 0) > 0 then Lit $ Is_Gt_One i d else F
		Fork		-> block_bw x qs (init c 0) d live vars nfqs show_p_t invs
		Join di		-> block_bw x qs (init c $ di) d live vars nfqs show_p_t invs
		Switch f s	-> case d of
						Just d' -> if Seq.elemIndexL d' (curr_colors c p) /= Nothing then block_bw x qs (init c 0) d live vars nfqs show_p_t invs else F
						Nothing -> Seq.foldrWithIndex (\n color -> disjunct $ if f color == (p == 0) then block_bw x qs (init c 0) (Just color) live vars nfqs show_p_t invs else F) F (Seq.index (colors c) 0)
		Merge		-> disjunct (if d == Nothing || Seq.elemIndexL (fromJust $ d) (Seq.index (colors c) 0) /= Nothing then {-trace (ide c ++ " select(3) up in") $-} conjunct (Lit $ Select i True) $ block_bw x qs (init c 0) d live vars nfqs show_p_t invs else F)
								(if d == Nothing || Seq.elemIndexL (fromJust $ d) (Seq.index (colors c) 1) /= Nothing then {-trace (ide c ++ " select(4) down in") $-} conjunct (Lit $ Select i False) $ block_bw x qs (init c 1) d live vars nfqs show_p_t invs else F)
		Function f s  -> case d of
							Just d' -> if Seq.elemIndexL d' (curr_colors c p) == Nothing then F else
										Seq.foldrWithIndex (\n color -> conjunct $ if f color == d' then block_bw x qs (init c 0) (Just color) live vars nfqs show_p_t invs else T) T (Seq.index (colors c) 0)
							Nothing -> if Seq.length (curr_colors c p) == 0 then F else block_bw x qs (init c 0) d live vars nfqs show_p_t invs
		Source injs s -> case d of
							Nothing -> fromBool $ Prelude.length injs /= 0
							Just d  -> fromBool $ d `elem` injs
		Automaton ss token ts -> case d of
									Just d' -> foldr (\(p',color') -> disjunct ( if any (\(s,s',e,f) -> e p' color' && f p' color' == Just (p,d')) ts then
																					block_bw x qs (init c p') d live vars nfqs show_p_t invs
																	  			 else
																					F))
													 F (ins_with_colors c)
	where trgt c p = case Seq.index (outs c) p of
						Just (Left i,p') -> (i,p')
	      init c p = case Seq.index (ins c) p of
						Just (Left i,p') -> (i,p')
	      curr_colors c p = let (i',p') = trgt c p in
								Seq.index (colors (Seq.index x i')) p'
	      ins_with_colors c = concat $ map (\p' -> map (\color -> (p',color)) (toList $ Seq.index (colors c) p')) [0..Seq.length (ins c)-1]



isSource :: Component p_t -> Bool
isSource c = case prim c of
				Source injs s -> True
				_ -> False

-- When established that a channel is live, this can be propagated trough the network.
-- Updates the given sequence of live components by adding new live ones.
propagate_liveness :: Eq p_t => Network p_t -> Int -> Int -> Seq.Seq Int -> Seq.Seq Int
propagate_liveness x i p live =
	case Seq.elemIndexL i live of
		Just n -> live
		Nothing -> let c = Seq.index x i in
					if not (isSource c) && Seq.length (init_colors c p) == 0 then live
					else
					  trace ((ide c) ++ " is live (FW).") $
					  let live' = (Seq.<|) i live;
						  live'' = 
						foldr (goto_out c) live'
						(case prim c of
							Queue n	    -> [0] 
							Merge       -> [0]
							Switch f s  -> Prelude.filter (\p -> Seq.length (trgt_colors c p) > 0) [0,1]
							ASwitch fs  -> Prelude.filter (\p -> Seq.length (trgt_colors c p) > 0) [0,1]
							Source inj s-> if inj /= [] then [0] else []
							Fork		-> [0,1]
							Join di		-> [0]
							Function f s-> [0] 
							_			-> []
						) in
						foldr (goto_in c) live''
						(case prim c of
							Join di		-> [other p]
							_			-> []
						)
	where
		goto_out c p' live = case (Seq.index (outs c) p') of
								Just (Left i,p) -> propagate_liveness x i p live 
		goto_in c p' live = case (Seq.index (ins c) p') of
								Just (Left i,p) -> propagate_liveness_bw x i p live 
		trgt c p = case Seq.index (outs c) p of
						Just (Left i,p') -> (i,p')
		trgt_colors c p = let (i',p') = trgt c p in
							Seq.index (colors (Seq.index x i')) p'
		init_colors c p = Seq.index (colors c) p

propagate_liveness_bw :: Eq p_t => Network p_t -> Int -> Int -> Seq.Seq Int -> Seq.Seq Int
propagate_liveness_bw x i p live =
	case Seq.elemIndexL i live of
		Just n -> live
		Nothing -> let c = Seq.index x i in
					if Seq.length (trgt_colors c p) == 0 then live
					else
					  trace ((ide c) ++ " is live (BW).") $
					  let live' = (Seq.<|) i live;
						  live'' = 
						foldr (goto_in c) live'
						(case prim c of
							Queue n	    -> [0] 
							Merge       -> Prelude.filter (\p -> Seq.length (init_colors c p) > 0) [0,1]
							Switch f s  -> [0]
							ASwitch fs  -> []
							Fork		-> [0]
							Join di		-> [0,1]
							Sink dead	-> [0]
							Function f s-> [0] 
							_			-> []
						) in
						foldr (goto_out c) live''
						(case prim c of
							Switch f s  -> if Seq.length (trgt_colors c $ other p) > 0 then [other $ p] else []
							Fork		-> [other p]
							_			-> []
						)
	where
		goto_out c p' live = case (Seq.index (outs c) p') of
								Just (Left i,p) -> propagate_liveness x i p live 
		goto_in c p' live = case (Seq.index (ins c) p') of
								Just (Left i,p) -> propagate_liveness_bw x i p live 
		trgt c p = case Seq.index (outs c) p of
						Just (Left i,p') -> (i,p')
		trgt_colors c p = let (i',p') = trgt c p in
							Seq.index (colors (Seq.index x i')) p'
		init_colors c p = Seq.index (colors c) p


