module Mesh_MI where

import Datastructures
import Data.Sequence (fromList)
import Control.Exception (assert)
import qualified Data.Sequence as Seq

-- build a merge with $n$ inputs from normal merges
n_merge n idx trgt = assert (n >= 2)
			map (\i -> (Merge, idx ++ show i, [Nothing,Nothing], (if i == n - 2 then [trgt] else [Just(idx ++ (show $ i + 1),0)])))
				[0..n-2]

data Color = GetX | FwdGetX | Data | PutX | WriteBackNack | WriteBackAck | Invalidate | ReplEvent | Load | InvEvent | ReplAck | DataRecd
  deriving (Show, Eq, Ord, Enum)
data Packet = Token | Packet {
	x :: Int,
	y :: Int,
	x_src :: Int,
	y_src :: Int,
	color :: Color
} deriving (Show, Eq, Ord)

allPackets = [Token] ++
			 map (\(x,y,x',y',c) -> Packet x y x' y' c) [(x,y,x',y',c) | x <- [0,1], y <- [0,1], x' <- [0,1], y' <- [0,1], c <- [GetX ..]]


show_p Token = "token"
show_p (Packet x y x' y' c) = "x" ++ show x ++ "_y" ++ show y ++ "xx" ++ show x' ++ "_yy" ++ show y' ++ "_c" ++ show c

isToken Token = True
isToken _ = False
isLoad Load = True
isLoad _ = False
isWBN (Packet x y x' y' WriteBackNack) = True
isWBN _ = False
isWBA (Packet x y x' y' WriteBackAck) = True
isWBA _ = False
isPutX (Packet x y x' y' PutX) = True
isPutX _ = False
isFwdGetX (Packet x y _ _ FwdGetX) x' y' = (x,y) == (x',y')
isFwdGetX _ _ _ = False
isInv (Packet x y x' y' Invalidate) = True
isInv _ = False
isData (Packet x y _ _ Data) x' y' = (x,y) == (x',y')
isData _ _ _ = False
isGetX (Packet x y x' y' GetX) = True
isGetX _ = False
isDataRecd (Packet x y x' y' DataRecd) = True
isDataRecd _ = False

intoToken d = Token



to_local x' y' = Switch (\d -> x d == x' && y d == y') ""
to_x     x' y' = Switch (\d -> x d /= x') ""
to_N     x' y' = Switch (\d -> y d < y') ""
to_W     x' y' = Switch (\d -> x d < x') ""


xy_node_p x y = [
			(Queue qsize,			"N",			[Nothing],			[Just("switchLN",0)]),
			(Queue qsize,			"S",			[Nothing],			[Just("switchLS",0)]),
			(Queue qsize,			"E",			[Nothing],			[Just("switchLEV",0)]),
			(Queue qsize,			"W",			[Nothing],			[Just("switchLWV",0)]),
			(to_local x y,			"switchLN",		[Nothing],			[Just("mergeO",0), Just("mergeN",0)]),
			(to_local x y,			"switchLS",		[Nothing],			[Just("mergeO",1), Just("mergeS",0)]),
			(to_local x y,			"switchLEV",	[Nothing],			[Just("mergeO",2), Just("switchEV",0)]),
			(to_local x y,			"switchLWV",	[Nothing],			[Just("mergeO",3), Just("switchWV",0)]),
			(to_x x y,				"switchEV",		[Nothing],			[Just("mergeE",1), Just("switchENS",0)]),
			(to_x x y,				"switchWV",		[Nothing],			[Just("mergeW",1), Just("switchWNS",0)]),
			(to_N x y,				"switchENS",	[Nothing],			[Just("mergeN",1), Just("mergeS",1)]),
			(to_N x y,				"switchWNS",	[Nothing],			[Just("mergeN",2), Just("mergeS",2)]),
			(to_local x y,			"switchIn",		[Nothing],			[Just("mergeO",4), Just("switchIn2",0)]),
			(to_x x y,				"switchIn2",	[Nothing],			[Just("switchWE",0), Just("switchNS",0)]),
			(to_W x y,				"switchWE",		[Nothing],			[Just("mergeW",0), Just("mergeE",0)]),
			(to_N x y,				"switchNS",		[Nothing],			[Just("mergeN",3), Just("mergeS",3)]),
			(Merge,					"mergeE",		[Nothing,Nothing],	[Just("outE", 0)]),
			(Merge,					"mergeW",		[Nothing,Nothing],	[Just("outW", 0)])
		]

natDecr :: Int -> Int
natDecr n = if n == 0 then 0 else n - 1

xy_node x y = Hierarchy (fromList $
							[Flat $ build $ xy_node_p x y]
							++
						    [Flat $ build $ n_merge 5 "mergeO" (Just("arrive",0))]
							++
						    [Flat $ build $ n_merge 4 "mergeN" (Just("outN",0))]
							++
						    [Flat $ build $ n_merge 4 "mergeS" (Just("outS",0))])
						(\n s p -> case (n,s) of
									(0,"mergeO") -> (1, "mergeO" ++ show (natDecr p), if p == 0 then 0 else 1)
									(0,"mergeN") -> (2, "mergeN" ++ show (natDecr p), if p == 0 then 0 else 1)
									(0,"mergeS") -> (3, "mergeS" ++ show (natDecr p), if p == 0 then 0 else 1)
									_ -> (n,s,p))




isReplEvent (Packet _ _ _ _ ReplEvent) = True
isReplEvent _ = False





dimx = 4
dimy = 4

dirx = 3
diry = 3

qsize = 24 -- With injection/arrival queuesizes should at least be equal to num of caches for DLF.


num_of_caches = dimx*dimy-1
dir_number 	  = dirx + dimx*diry
cache_numbers = filter (/= dir_number) [0..dimx*dimy-1]

insert_load_repl x y = Source [Packet x y x y Load,Packet x y x y ReplEvent] ""

cache x y = 
	let cache_name = "cache" in
	[
			(Automaton ["I","IM", "M", "MI", "MII"] Token [
							(0,1,\p d-> p==0 && d == Packet x y x y Load,					\p d -> Just (0, Packet dirx diry x y GetX)),
							(1,2,\p d-> p==1 && isData d x y ,								\p d -> Just (0, Packet dirx diry x y DataRecd)),
							(2,3,\p d-> p==1 && d == Packet x y dirx diry Invalidate,		\p (Packet x y x_src y_src Invalidate) -> Just (0, Packet x_src y_src x y PutX)),
							(2,4,\p d-> p==0 && d == Packet x y x y ReplEvent,				\p d -> Just (0, Packet dirx diry x y ReplEvent)),
							(3,0,\p d-> p==1 && d == Packet x y dirx diry WriteBackAck,		\p d -> Nothing),
							(2,0,\p d-> p==1 && isFwdGetX d x y, 							\p (Packet x y x_src y_src FwdGetX) -> Just (0, Packet x_src y_src x y Data)),
							(4,0,\p d-> p==1 && d == Packet x y dirx diry ReplAck,			\p d -> Nothing),
							(4,2,\p d-> p==1 && d == Packet x y dirx diry WriteBackNack,	\p d -> Nothing)
									],
						cache_name, [Nothing,Nothing], [Just ("injectQ", 0)]),
			(insert_load_repl x y, "src", [], [Just(cache_name,0)]),
		--	(Function id "", "injectQ", [Nothing], [Just("inject", 0)]),
		--	(Function id "", "arriveQ", [Nothing], [Just(cache_name,1)])
			(Queue qsize, "injectQ", [Nothing], [Just("inject", 0)]),
			(Queue qsize, "arriveQ", [Nothing], [Just(cache_name,1)])
	]

insert_invevent = Source (map (\n -> let (x,y) = (n `mod` dimx, n `div` dimx) in
									Packet dirx diry x y InvEvent)
								 cache_numbers) ""
insert_token = Source [Token] ""

directory =
	let sI	 = 0;
		sIM	 = 1;
		sMI	 = 2;
		sMF	 = 3 in
		[
			(Automaton	(["I", "IM", "MI", "MF"] ++ (concatMap (\s -> map (\cache -> s ++ show cache) cache_numbers) ["M"]))
					  	Token
						((concatMap (\(n,cn) -> let (x,y) = (cn `mod` dimx, cn `div` dimx);
													sM	 = 4+n in
							[
								(sIM,sM,	\p d -> p==1 && d == Packet dirx diry x y DataRecd,	\p d -> Nothing),
								(sM,sI,		\p d -> p==1 && d == Packet dirx diry x y ReplEvent,\p (Packet x y x_src y_src ReplEvent) -> Just (0, Packet x_src y_src dirx diry ReplAck)),
								(sM,sMI,	\p d -> p==0 && isToken d,							\p d -> Just (0, Packet x y dirx diry Invalidate)),
								(sM,sMF,	\p d -> p==1 && isGetX d,							\p (Packet dirx diry x_src y_src GetX) -> Just (0, Packet x y x_src y_src FwdGetX)),
								(sMF,sM,	\p d -> p==1 && d == Packet dirx diry x y DataRecd, \p d -> Nothing)
							]
							)
									(zip [0..] cache_numbers))
							++
								[(sI,sIM,	\p d -> p==1 && isGetX d,	 	\p (Packet x y x_src y_src GetX) -> Just (0, Packet x_src y_src dirx diry Data)),
								 (sMI,sI,	\p d -> p==1 && isPutX d,		\p (Packet x y x_src y_src PutX) -> Just (0, Packet x_src y_src dirx diry WriteBackAck)),
								 (sIM,sIM,	\p d -> p==1 && isReplEvent d,	\p (Packet x y x_src y_src ReplEvent) -> Just (0, Packet x_src y_src dirx diry WriteBackNack)),
								 (sMI,sMI,	\p d -> p==1 && isReplEvent d,	\p (Packet x y x_src y_src ReplEvent) -> Just (0, Packet x_src y_src dirx diry WriteBackNack)),
								 (sMF,sMF,	\p d -> p==1 && isReplEvent d,	\p (Packet x y x_src y_src ReplEvent) -> Just (0, Packet x_src y_src dirx diry WriteBackNack))
								 ]),
						"dir", [Nothing,Nothing], [Just ("injectQ", 0)]),
			(insert_token, "src", [], [Just("dir",0)]),
		--	(Function id "", "injectQ", [Nothing], [Just("inject", 0)]),
		--	(Function id "", "arriveQ", [Nothing], [Just("dir",1)])
			(Queue qsize, "injectQ", [Nothing], [Just("inject", 0)]),
			(Queue qsize, "arriveQ", [Nothing], [Just("dir",1)])
		]



hnetwork :: HNetwork Packet
hnetwork = Hierarchy (Seq.fromList $
					  (map (\n -> xy_node (n `mod` dimx) (n `div` dimx)) [0..dimx*dimy-1])
						++
					  (map (\n -> if n == dir_number then
									Flat $ build $ directory
								  else
									Flat $ build $ cache (n `mod` dimx) (n `div` dimx)) [0..dimx*dimy-1])
						)
					  (\n s p -> case s of
							   "outE" -> if n `mod` dimx == dimx-1 then (n,"snk",0) else (n+1, "E",0)
							   "outW" -> if n `mod` dimx == 0 then (n,"snk",0) else (n-1, "W",0)
							   "outN" -> if n `div` dimx == 0 then (n,"snk",0) else (n-dimx, "N",0)
							   "outS" -> if n `div` dimx == dimy-1 then (n,"snk",0) else (n+dimx, "S",0)
							   "arrive" -> (n+dimx*dimy, "arriveQ",0)
							   "inject" -> (n-dimx*dimy, "switchIn",0)
							   _      -> (n,s,p))



