module Torus_Uniring where

import Datastructures
import Data.Sequence (fromList)
import Control.Exception (assert)
import qualified Data.Sequence as Seq

n_merge n idx trgt = assert (n >= 2)
			map (\i -> (Merge, idx ++ show i, [Nothing,Nothing], (if i == n - 2 then [trgt] else [Just(idx ++ (show $ i + 1),0)])))
				[0..n-2]



data Packet = Token | Req Int Int | Data Int Int
  deriving (Show, Eq, Ord)

show_p Token = "token"
show_p (Req n n_src) = "Req_n" ++ show n ++ "_nn" ++ show n_src
show_p (Data x y) = "Data_x" ++ show x ++ "_y" ++ show y 

isToken Token = True
isToken _ = False
isData (Data x y) = True
isData _ = False
isReq (Req n n') = True
isReq _ = False

dimx = 5
dimy = 5
qsize = 10

to_local x' y' = Switch (\(Data x_d y_d) -> x_d == x' && y_d == y') ""
to_x     x' y' = Switch (\(Data x_d y_d) -> x_d /= x') ""
to_N     x' y' = Switch (\(Data x_d y_d) -> (y' + dimy - y_d) `mod` dimy <= dimy `div` 2) ""
to_W     x' y' = Switch (\(Data x_d y_d) -> (x' + dimx - x_d) `mod` dimx <= dimx `div` 2) ""


xy_node_p x y = [
			(Queue qsize,			"N",			[Nothing],			[Just("switchLN",0)]),
			(Queue qsize,			"S",			[Nothing],			[Just("switchLS",0)]),
			(Queue qsize,			"E",			[Nothing],			[Just("switchLEV",0)]),
			(Queue qsize,			"W",			[Nothing],			[Just("switchLWV",0)]),
			(to_local x y,			"switchLN",		[Nothing],			[Just("mergeO",0), Just("mergeN",0)]),
			(to_local x y,			"switchLS",		[Nothing],			[Just("mergeO",1), Just("mergeS",0)]),
			(to_local x y,			"switchLEV",	[Nothing],			[Just("mergeO",2), Just("switchEV",0)]),
			(to_local x y,			"switchLWV",	[Nothing],			[Just("mergeO",3), Just("switchWV",0)]),
			(to_x x y,				"switchEV",		[Nothing],			[Just("mergeE",1), Just("switchENS",0)]),
			(to_x x y,				"switchWV",		[Nothing],			[Just("mergeW",1), Just("switchWNS",0)]),
			(to_N x y,				"switchENS",	[Nothing],			[Just("mergeN",1), Just("mergeS",1)]),
			(to_N x y,				"switchWNS",	[Nothing],			[Just("mergeN",2), Just("mergeS",2)]),
			(to_local x y,			"switchIn",		[Nothing],			[Just("mergeO",4), Just("switchIn2",0)]),
			(to_x x y,				"switchIn2",	[Nothing],			[Just("switchWE",0), Just("switchNS",0)]),
			(to_W x y,				"switchWE",		[Nothing],			[Just("mergeW",0), Just("mergeE",0)]),
			(to_N x y,				"switchNS",		[Nothing],			[Just("mergeN",3), Just("mergeS",3)]),
			(Merge,					"mergeE",		[Nothing,Nothing],	[Just("outE", 0)]),
			(Merge,					"mergeW",		[Nothing,Nothing],	[Just("outW", 0)])
		]

natDecr :: Int -> Int
natDecr n = if n == 0 then 0 else n - 1

xy_node x y = Hierarchy (fromList $
							[Flat $ build $ xy_node_p x y]
							++
						    [Flat $ build $ n_merge 5 "mergeO" (Just("arrive",0))]
							++
						    [Flat $ build $ n_merge 4 "mergeN" (Just("outN",0))]
							++
						    [Flat $ build $ n_merge 4 "mergeS" (Just("outS",0))])
						(\n s p -> case (n,s) of
									(0,"mergeO") -> (1, "mergeO" ++ show (natDecr p), if p == 0 then 0 else 1)
									(0,"mergeN") -> (2, "mergeN" ++ show (natDecr p), if p == 0 then 0 else 1)
									(0,"mergeS") -> (3, "mergeS" ++ show (natDecr p), if p == 0 then 0 else 1)
									_ -> (n,s,p))

insert_token x y = Source [Token] ""
cache x y = [
			(Automaton ["I","IM", "M"] Token [
							(0,1,\p d-> p==0 && isToken d,	
														\p d -> Just (0, Req (((x+(dimx*y))+1)`mod`(dimx*dimy)) (x+(dimx*y)))),
							(0,0,\p d-> p==1 && (case d of Req n n' -> n == x+(dimx*y) && n' /= x+(dimx*y); _ -> False),	
														\p (Req n n') -> Just (0, Req ((n+1)`mod`(dimx*dimy)) n')),
							(1,1,\p d-> p==1 && (case d of Req n n' -> n == x+(dimx*y) && n' /= x+(dimx*y); _ -> False),
														\p (Req n n') -> Just (0, Req ((n+1)`mod`(dimx*dimy)) n')),
							(1,2,\p d-> (p==2 && isData d) || (p==1 && (case d of Req n n' -> n' == x+(dimx*y); _ -> False)),
														\p d -> Nothing),
							(2,0,\p d-> p==1 && (case d of Req n n' -> n == x+(dimx*y) && n' /= x+(dimx*y); _ -> False),
														\p (Req n n') -> Just (0, Req ((n+1)`mod`(dimx*dimy)) n')),
							(2,0,\p d-> p==0 && isToken d,
														\p d -> Nothing),
							(2,0,\p d-> p==1 && (case d of Req n n' -> n' /= x+dimx*y; _ -> False),	
														\p (Req n n') -> Just (1, Data (n' `mod` dimx) (n' `div` dimx)))
												],
						"cache", [Nothing,Nothing,Nothing], [Just ("ring" ++ show (y*dimx+x),0), Just ("injectQ", 0)]),
			(insert_token x y, "src", [], [Just("cache",0)]),
			(Queue qsize, "injectQ", [Nothing], [Just("inject", 0)]),
			(Queue 2, 	  "ring" ++ show (y*dimx+x), [Nothing], [Just("nextcache", 1)]),
			(Queue qsize, "arriveQ", [Nothing], [Just("cache",2)])
	]





hnetwork :: HNetwork Packet
hnetwork = Hierarchy (Seq.fromList $
					  (map (\n -> xy_node (n `mod` dimx) (n `div` dimx)) [0..dimx*dimy-1])
						++
					  (map (\n -> Flat $ build $ cache (n `mod` dimx) (n `div` dimx)) [0..dimx*dimy-1])
						)
					  (\n s p -> case s of
							   "outE" -> (if n `mod` dimx == dimx-1 then dimx*(n`div`dimx) else n+1, "E",0)
							   "outW" -> (if n `mod` dimx == 0 then dimx*(n`div`dimx) + dimx-1 else n-1, "W",0)
							   "outN" -> (if n `div` dimx == 0 then (n `mod` dimx) + dimx*(dimy-1) else n-dimx, "N",0)
							   "outS" -> (if n `div` dimx == dimy-1 then n `mod` dimx else n+dimx, "S",0)
							   "arrive" -> (n+dimx*dimy, "arriveQ",0)
							   "nextcache" -> (if n < 2*dimx*dimy-1 then n+1 else dimx*dimy, "cache",1)
							   "inject" -> (n-dimx*dimy, "switchIn",0)
							   _      -> (n,s,p))








