module Main where
import Datastructures
import Analysis
import DeadlockDetection
import SMT
import Data.ByteString.Lazy (hGetContents)
import System.Environment
import Control.Exception (assert)
import qualified Data.Sequence as Seq
import Debug.Trace
import Control.Monad
import qualified Data.Map as Map
import qualified Data.Set as Set
import qualified Data.Foldable as Fold
import Data.Data
import Data.Maybe
import System.IO
import System.Process
import System.Environment
import System.Exit
import qualified Control.Monad.Parallel as Par
import System.IO.Temp
import Control.Parallel.Strategies
import Data.List.Split (chunksOf)
import Control.Applicative ((<$>))
import GHC.Conc (numCapabilities)



-- Import one of the following examples:
import Cross_Deadlock
--import Cross_Deadlock_VC
--import Mesh_MI
--import Torus_Uniring



-- EXAMPLES:
-- A simplified version of Fig 6 from Brayton. Has a deadlock, only due to the merge.
{-
network = [
			(Source [0::Int] "", 		"src", 	[],					[Just("mrg",0)]),
			(Merge,						"mrg",	[Nothing, Nothing],	[Just("sw0",0)]),
			(Switch (\d -> d == 0) "",  "sw0",	[Nothing],			[Just("q0",0),Just("q1",0)]),
			(Queue 2,	 				"q0",	[Nothing],			[Just("f",0)]),
			(Queue 2,	 				"q1",	[Nothing],			[Just("snk",0)]),
			(Function (\d -> 1) "",		"f",	[Nothing],			[Just("mrg",1)]),
			(Sink False, 				"snk",	[Nothing],			[])
		  ]
-}
-- Two queues between a fork/join. The 0-packets are routed out of the fork/join, the 1's stay within.
-- Yields a color-specific invariant
{-
network = [
			(Source [0,1] "",	 		"src", 	[],					[Just("f1",0::Int)]),
			(Fork,						"f1",	[Nothing],			[Just ("q0",0),Just ("q1",0)]),
			(Queue 2,	 				"q0",	[Nothing],			[Just("sw0",0)]),
			(Queue 2,	 				"q1",	[Nothing],			[Just("sw1",0)]),
			(Switch (\d -> d == 0) "",  "sw0",	[Nothing],			[Just("snk0",0),Just("j1",0)]),
			(Switch (\d -> d == 1) "",  "sw1",	[Nothing],			[Just("j1",1),Just("snk2",0)]),
			(Join 0,					"j1",	[Nothing, Nothing],	[Just("snk1",0)]),
			(Sink False,				"snk0",	[Nothing],			[]),
			(Sink False, 				"snk1",	[Nothing],			[]),
			(Sink False, 				"snk2",	[Nothing],			[])
		  ]
-- Two queues between a fork/join. The upper 0-packets are routed out of the fork/join, the rest stays within.
network = [
			(Source [0,1] "",	 		"src", 	[],					[Just("f1",0::Int)]),
			(Fork,						"f1",	[Nothing],			[Just ("q0",0),Just ("q1",0)]),
			(Queue 2,	 				"q0",	[Nothing],			[Just("sw",0)]),
			(Queue 2,	 				"q1",	[Nothing],			[Just("j1",1)]),
			(Switch (\d -> d == 0) "",  "sw",	[Nothing],			[Just("snk0",0),Just("j1",0)]),
			(Join 0,					"j1",	[Nothing, Nothing],	[Just("snk1",0)]),
			(Sink False, 				"snk0",	[Nothing],			[]),
			(Sink False, 				"snk1",	[Nothing],			[])
		  ]
-- (src, fork, 2/1 queues up/down, join, sink) 
network = [
			(Source [0,1] "",	"src", 	[],								[Just("f1",0::Int)]),
			(Fork,				"f1",	[Just("src",0)],				[Nothing,Just ("q3",0)]),
			(Queue 2,		 	"q1",	[Just("f1", 0)],				[Just("q2",0)]),
			(Queue 2,		 	"q2",	[Just("q1", 0)],				[Just("j1",0)]),
			(Queue 2,		 	"q3",	[Nothing],						[Nothing]),
			(Join 1,			"j1",	[Just("q2", 0), Just("q3",0)],	[Just("snk",0)]),
			(Sink False, 		"snk",	[Just("j1", 0)],				[])
		  ]
-}









check_feasibility i x qs vars show_p_t invs solver f = 
	case f of
		T -> return True
		F -> return False
		_ -> withTempFile "." ("t"++show i++".smt2") $ \file h ->
			 	do 
					let ret  = simplify_subsumptions $ simplify_remove_selects $ f
					hPutStrLn h $ "(set-logic QF_LIA)\n" ++ invs
					let (smt',qs',vars') = export_formula_to_SMT x qs vars show_p_t Nothing f
					hPutStrLn h $ smt'
					hPutStrLn h $ "(check-sat)\n(get-model)"
					() <- hClose h
					(exit,solver_output,err) <- (uncurry readProcessWithExitCode) (solver file) [] 
					case parseSMTOutput solver_output of
						Left err -> do
										print err
										h <- openFile "temp.smt2" WriteMode
										hPutStrLn h $ "(set-logic QF_LIA)\n" ++ invs
										hPutStrLn h $ "(set-logic QF_LIA)\n" ++ invs ++ smt' ++ "(check-sat)\n(get-model)"
										hClose h
										exitFailure
						Right Nothing -> return False
						Right (Just model) -> return True





-- | Like 'Control.Monad.filterM', but applying the function to the individual list items in parallel (in chunks).
par_filterM :: (Functor m, Par.MonadParallel m) => (a -> m Bool) -> [a] -> m [a]
par_filterM f list = concat <$> Par.mapM (filterM f) (chunksOf (length list `quot` numCapabilities) list)
						

cvc4,z3 :: String -> (String, [String])
cvc4 file = ("cvc4", ["-m",file])
z3	 file = ("z3", [file])

add_formulas_to_invariants x show_p_t =
	foldr (\f (invs,qs,vars) -> let (invs',qs',vars') = export_formula_to_SMT x qs vars show_p_t Nothing f in (invs ++ invs',qs',vars'))

main :: IO ()
main = do
		args <- getArgs
		let solver = if "z3" `elem` args then z3 else cvc4
		putStrLn $ "Using " ++ show numCapabilities ++ " cores."




		-- Use on of the following two lines:
		let (x,show_p_t) = (derive_colors $ set_pointers $ fill_blanks $ flatten hnetwork, show_p)
		--let (x,show_p_t) = (derive_colors $ set_pointers $ fill_blanks $ build $ network, show_p)



		-- First, generate invariants the Intel way
		let (invs',qs',vars') = export_invariants_to_smt x show_p_t


		-- Maybe, assert some invariants manually
		let minvs = []
					-- [manually_assert x Nothing $ ("H" ++ show (dir_number + dimx*dimy) ++ "_dir", "I") : map (\n -> ("H" ++ show (n+dimx*dimy) ++ "_cache", "IM")) cache_numbers]
		let (invs,qs,vars) = add_formulas_to_invariants x show_p_t (invs',qs',vars') minvs

		putStrLn $ "Num of comb. primitives: " ++ show (Seq.length x - dimx*dimy - numOfQueues x) ++ "; num of queues: " ++ show (numOfQueues x) ++ "; num of automata: " ++ show (dimx*dimy)

		-- Thirdly, figure out if there are queues that are never full.
		nfqs <- par_filterM (\i -> case prim $ Seq.index x i of
										Queue s -> do
													feasible <- check_feasibility i x qs' vars' show_p_t invs' solver (Lit $ Is_Full i)
													print (i, ide (Seq.index x i), feasible)
													return $ not feasible
										_ -> return False)
						[0..Seq.length x -1]
		if nfqs /= [] then
			putStrLn("List of queues that cannot be full: " ++ show nfqs)
		else
			return ()

		-- Write invariants to a file
		writeFile "invariants.smt2" $ "(set-logic QF_LIA)\n" ++ invs
		putStrLn "Derived color information and invariants, written file \"invariants.smt\"."


		-- Thirdly, start deadlock detection
		(live,dl) <- detectDeadlocks x show_p_t solver invs qs vars nfqs
		return ()
	where
		numOfQueues x = Seq.length (Seq.filter (\c -> case prim c of Queue s -> True; _ -> False) x)
		detectDeadlocks x show_p_t solver invs qs vars nfqs =
			Fold.foldrM (\i (live,dl) ->
							(if not dl then
								case prim $ Seq.index x i of
									Source injs s -> if injs /= [] then do
														putStrLn ("Starting deadlock detection for source " ++ (ide $ Seq.index x i))
														detectDeadlock x show_p_t solver live invs qs vars nfqs i
													 else do
														putStrLn $ "Source " ++ ide (Seq.index x i) ++ " is idle."
														return (live,dl)
									_ -> return (live,dl)
							else
								return (live,dl)))
					(Seq.empty, False)
					[0..Seq.length x - 1]
		-- Determine the block formula of the source. Export it to an SMT solver.
		-- Based on the result, update the set of live components or not.
		detectDeadlock x show_p_t solver live invs qs vars nfqs src = 
				do
					let ret  = unfold_formula x qs live vars nfqs show_p_t invs (Lit $ Block src 0 Nothing) Map.empty
					--() <- traceShow ret $ return ()
					let file = "deadlock_" ++ ide (Seq.index x src) ++ ".smt2"
					h <- openFile file WriteMode
					hPutStrLn h $ "(set-logic QF_LIA)\n" ++ invs
					mapM_ (\(bi, f) -> hPutStrLn h $ export_bi_var_to_smt x show_p_t bi) (Map.toList ret)
					foldM_ (\(qs,vars) (bi, f) -> do 
														let (smt',qs',vars') = export_formula_to_SMT x qs vars show_p_t (Just bi) f
														hPutStrLn h $ smt'
														return (qs',vars'))
						   (qs,vars) (Map.toList ret)
					hPutStrLn h $ "(assert " ++ bi_to_name x show_p_t (Block src 0 Nothing) ++ ")"
					hPutStrLn h $ "(check-sat)\n(get-model)"
					hClose h
					putStrLn $ "Starting solver on file: " ++ file
					(exit,solver_output,err) <- (uncurry readProcessWithExitCode) (solver file) []
					case parseSMTOutput solver_output of
							Left err -> do
											print err
											exitFailure
							Right Nothing -> do
												putStrLn $ "Source " ++ ide (Seq.index x src) ++ " is live."
												return $ (propagate_liveness x src 0 live,False)
							Right (Just model) -> do
													putStrLn $ "Source " ++ ide (Seq.index x src) ++ " has a possible deadlock:\n" ++ showModel model
													return (live,True)


